﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace PojeClient {
    // AIを作る際にあると便利な関数　
    class AITools {

        // 盤 board で手 actionを打った場合に，どういう攻撃になるか．      
        // updateSrcBoard = true の場合，引数のsrcBoard自体をいじる．falseの場合はコピーしてからいじる（安全）
        public static AITool_Result calcTotalGain(Board board, Action action, bool updateSrcBoard) {
            if (updateSrcBoard == false) board = new Board(board);
            List<Point> drops = board.transit1_ActionDrop(action);
            int colorVanishCount = board.transit2_calcVanishPoints(drops);
            if (colorVanishCount == 0) {
                // その後のboardを使うことがあるので，配珠をずらす（次が読めるようにする）．
                board.setNextColors(new int[2] { 0, 0 });
                return new AITool_Result(false, 0, 0, action, board);
            }

            // 消えなくなるまでループする
            int chainCount = 0;
            int totalGain = 0;
            int jamVanishCount = 0;
            do {
                chainCount++;
                jamVanishCount += board.getVanishPoints().Count - colorVanishCount; // 消える箇所の全体数 - 消える色珠数 = 消えるJamの数
                totalGain += JamInfo.getGainJamCount(chainCount, colorVanishCount);
                board.transit3_Vanish();
                drops = board.transit1_ChainDrop();  //連鎖中は行動なし，邪魔珠落下もなし
                colorVanishCount = board.transit2_calcVanishPoints(drops);
            } while (colorVanishCount > 0);

            // その後のboardを使うことがあるので，配珠をずらす（次が読めるようにする）．
            board.setNextColors(new int[2] { 0, 0 });
            return new AITool_Result(true, totalGain, jamVanishCount, action, board);
        }


        // あるターン数の中で最大の攻撃ができる手（の一手目）を選ぶ．
        // この結果は，その盤（とnexts）が持っている，攻撃ポテンシャルの一つ．
        public static AITool_Result calcBestAttackAction(Board board, List<Action> actions, int restTurn) {
            AITool_Result bestResult = null;

            // このターンで発動しなければいけない場合
            if (restTurn == 1) {
                foreach (Action a in actions) {
                    AITool_Result result = calcTotalGain(board, a, false);
                    if (bestResult == null) {
                        bestResult = result;
                    } else if (result.Gain > bestResult.Gain) {
                        bestResult = result;
                    }
                }
            } else {
                // まだターンに余裕がある場合は，発動しない手についてもその先を読む（発動するならするでよい）
                foreach (Action a in actions) {
                    AITool_Result result = calcTotalGain(board, a, false);

                    // 発動する手なら，result がその最終価値．　そうでないなら，次の局面での calcBestAttackActionが最終価値．
                    if (result.IsVanish == false) {
                        // その時点で打つ手がない場合候補にならない
                        if (result.NextBoard.createPossibleActions().Count == 0) continue;
                        // 手があるならそのポテンシャルを評価
                        result = calcBestAttackAction(result.NextBoard, result.NextBoard.getPossibleActions(), restTurn - 1);
                        // 選んだ手は a なのでそれを格納
                        result.Action = a;
                    }

                    // そのうえで，最善のものを比較格納
                    if (bestResult == null) {
                        bestResult = result;
                    } else if (result.Gain > bestResult.Gain) {
                        bestResult = result;
                    }
                }
            }

            return bestResult;



        }





        // 邪魔珠がいつ何個降ってくるか，現在相手が連鎖中の場合も含めて確定させる
        public static JamInfo calcOpponentsAttack(Board srcBoard, JamInfo srcJam, int turn) {
            int chainCount = srcBoard.getChainCount();
            // 相手が連鎖中でない場合は，そのままでよい
            if (chainCount == 0) return srcJam;

            JamInfo jam = new JamInfo(srcJam);
            Board board = new Board(srcBoard);

            // 消えなくなるまでループする
            while (true) {
                List<Point> drops = board.transit1_ChainDrop();  //連鎖中は行動なし，邪魔珠落下もなし
                int colorVanishCount = board.transit2_calcVanishPoints(drops);
                if (colorVanishCount == 0) {
                    // ターンの確定
                    jam.setJamTurn(turn+2);
                    break;
                }
                chainCount++;
                jam.addJam(JamInfo.getGainJamCount(chainCount, colorVanishCount));
                board.transit3_Vanish();
                turn++;
            }
            return jam;
        }


        // その手で落としたときに，周囲に同じ色があるか．弱いAIが連鎖しやすくするためのもの
        public static int calcConnectCount(Board srcBoard, Action action) {
            int count = 0;
            Point[] p = action.getDropPoints();
            for (int i = 0; i < 2; i++) {
                if (p[i].Y <= Settings.BoardHeight) {
                    for (int r = 0; r < 4; r++) {
                        if (srcBoard.getStone(p[i].X + Settings.RotateX[r], p[i].Y + Settings.RotateY[r]) == action.getColor(i)) count++;
                    }
                }
            }
            return count;
        }


        // 盤上，残り何マス空いてるか．緊急度の目安．
        public static int getRestCount(Board board) {
            int rest = 0;
            for (int x = 1; x <= Settings.BoardWidth; x++) {
                rest += Settings.BoardHeight - board.getHeight(x);
            }
            return rest;
        }










    }


    // いろんなものを返したい場合の構造体
    class AITool_Result {
        public bool IsVanish;
        public int Gain;
        public int JamVanishCount;
        public Action Action;
        public Board NextBoard;

        public AITool_Result(bool isVanish, int gain, int jamVanishCount, Action a, Board b) {
            IsVanish = isVanish;
            Gain = gain;
            JamVanishCount = jamVanishCount;
            Action = a;
            NextBoard = b;
        }
    }



}
